@@ -1,4 +1,5 @@
 """ Various utilties """
+from __future__ import print_function
 
 import datetime
 import re
@@ -13,6 +14,9 @@ import dateutil.tz
 import pytz
 import six
 
+from distutils.version import LooseVersion
+
+
 DATE_FORMAT = '%Y%m%dT%H%M%SZ'
 
 
@@ -31,7 +35,6 @@ encode_replacements_experimental = OrderedDict([
     ('"', '&dquot;'),
     ('[', '&open;'),
     (']', '&close;'),
-    ('\n', ' '),
 ])
 
 decode_replacements = OrderedDict([
@@ -40,14 +43,18 @@ decode_replacements = OrderedDict([
 ])
 
 logical_replacements = OrderedDict([
+    ('?', '\\?'),
+    ('+', '\\+'),
     ('(', '\\('),
     (')', '\\)'),
     ('[', '\\['),
     (']', '\\]'),
+    ('{', '\\{'),
+    ('}', '\\}'),
 ])
 
 
-def encode_task_value(value, query=False):
+def encode_task_value(key, value, query=False):
     if value is None:
         value = ''
     elif isinstance(value, datetime.datetime):
@@ -64,6 +71,9 @@ def encode_task_value(value, query=False):
             # In some contexts, parentheses are interpreted for use in
             # logical expressions.  They must *sometimes* be escaped.
             for left, right in six.iteritems(logical_replacements):
+                # don't replace '?' if this is an exact match
+                if left == '?' and '.' not in key:
+                    continue
                 value = value.replace(left, right)
         else:
             for unsafe, safe in six.iteritems(
@@ -75,15 +85,35 @@ def encode_task_value(value, query=False):
     return value
 
 
-def encode_query(value):
+def encode_query(value, version, query=True):
     args = []
-    for k, v in six.iteritems(value):
-        args.append(
-            '%s:\"%s\"' % (
-                k,
-                encode_task_value(v, query=True)
+
+    if isinstance(value, dict):
+        value = six.iteritems(value)
+
+    for k, v in value:
+        if isinstance(v, list):
+            args.append(
+                "( %s )" % (" %s " % k).join([
+                    encode_query([item], version, query=False)[0] for item in v
+                ])
             )
-        )
+        else:
+            if k.endswith(".is") and version >= LooseVersion('2.4'):
+                args.append(
+                    '%s == "%s"' % (
+                        k[:-3],
+                        encode_task_value(k, v, query=query)
+                    )
+                )
+            else:
+                args.append(
+                    '%s:%s' % (
+                        k,
+                        encode_task_value(k, v, query=query)
+                    )
+                )
+
     return args
 
 
@@ -101,13 +131,13 @@ def encode_task_experimental(task):
     if 'tags' in task:
         task['tags'] = ','.join(task['tags'])
     for k in task:
-        task[k] = encode_task_value(task[k])
+        task[k] = encode_task_value(k, task[k])
 
     # Then, format it as a string
-    return "%s\n" % " ".join([
-        "%s:\"%s\"" % (k, v)
+    return [
+        "%s:\"%s\"" % (k, v) if v else "%s:" % (k, )
         for k, v in sorted(task.items(), key=itemgetter(0))
-    ])
+    ]
 
 
 def encode_task(task):
@@ -200,3 +230,34 @@ def annotation_list_to_comparison_map(annotations):
         comparable = make_annotation_comparable(annotation)
         mapping[comparable] = annotation
     return mapping
+
+
+def convert_dict_to_override_args(config, prefix=''):
+    """ Converts a dictionary of override arguments into CLI arguments.
+
+    * Converts leaf nodes into dot paths of key names leading to the leaf
+      node.
+    * Does not include paths to leaf nodes not being non-dictionary type.
+
+    See `taskw.test.test_utils.TestUtils.test_convert_dict_to_override_args`
+    for details.
+
+    """
+    args = []
+    for k, v in six.iteritems(config):
+        if isinstance(v, dict):
+            args.extend(
+                convert_dict_to_override_args(
+                    v,
+                    prefix='.'.join([
+                        prefix,
+                        k,
+                    ]) if prefix else k
+                )
+            )
+        else:
+            v = six.text_type(v)
+            left = 'rc' + (('.' + prefix) if prefix else '') + '.' + k
+            right = v if ' ' not in v else '"%s"' % v
+            args.append('='.join([left, right]))
+    return args
