@@ -9,32 +9,37 @@ default TaskWarrior class.  If not, then the default TaskWarrior class will
 fall back to the older TaskWarriorDirect implementation.
 
 """
-
 import abc
 import codecs
 import copy
 from distutils.version import LooseVersion
+import logging
 import os
-import re
-import sys
 import time
 import uuid
 import subprocess
 import json
-import pprint
-
-import taskw.utils
-from taskw.exceptions import TaskwarriorError
 
 import six
 from six import with_metaclass
 from six.moves import filter
 from six.moves import map
-from six.moves import zip
+
+import taskw.utils
+from taskw.exceptions import TaskwarriorError
+from taskw.task import Task
+from taskw.taskrc import TaskRc
+
+
+logger = logging.getLogger(__name__)
 
 
 open = lambda fname, mode: codecs.open(fname, mode, "utf-8")
 
+# Location of configuration file: either specified by TASKRC environment
+# variable, or ~/.taskrc (default).
+TASKRC = os.getenv("TASKRC", "~/.taskrc")
+
 
 class TaskWarriorBase(with_metaclass(abc.ABCMeta, object)):
     """ The task warrior
@@ -43,9 +48,22 @@ class TaskWarriorBase(with_metaclass(abc.ABCMeta, object)):
     with a taskwarrior database.
     """
 
-    def __init__(self, config_filename="~/.taskrc"):
+    def __init__(
+        self,
+        config_filename=TASKRC,
+        config_overrides=None,
+        marshal=False
+    ):
         self.config_filename = config_filename
         self.config = TaskWarriorBase.load_config(config_filename)
+        if marshal:
+            raise NotImplementedError(
+                "You must use TaskWarriorShellout to use 'marshal'"
+            )
+        if config_overrides:
+            raise NotImplementedError(
+                "You must use TaskWarriorShellout to use 'config_overrides'"
+            )
 
     def _stub_task(self, description, tags=None, **kw):
         """ Given a description, stub out a task dict. """
@@ -83,7 +101,7 @@ class TaskWarriorBase(with_metaclass(abc.ABCMeta, object)):
                 else:
                     annotations.append(v)
 
-        for key in task.keys():
+        for key in list(task.keys()):
             if key.startswith('annotation_'):
                 annotations.append(task[key])
                 del(task[key])
@@ -101,7 +119,7 @@ class TaskWarriorBase(with_metaclass(abc.ABCMeta, object)):
 
         By default, the 'all' command is run.
 
-        >>> w = Warrior()
+        >>> w = TaskWarrior()
         >>> tasks = w.load_tasks()
         >>> tasks.keys()
         ['completed', 'pending']
@@ -145,7 +163,7 @@ class TaskWarriorBase(with_metaclass(abc.ABCMeta, object)):
         return filtered
 
     @classmethod
-    def load_config(self, config_filename="~/.taskrc"):
+    def load_config(cls, config_filename=TASKRC, overrides=None):
         """ Load ~/.taskrc into a python dict
 
         >>> config = TaskWarrior.load_config()
@@ -155,39 +173,7 @@ class TaskWarriorBase(with_metaclass(abc.ABCMeta, object)):
         'yes'
 
         """
-
-        with open(os.path.expanduser(config_filename), 'r') as f:
-            lines = f.readlines()
-
-        _usable = lambda l: not(l.startswith('#') or l.strip() == '')
-        lines = filter(_usable, lines)
-
-        def _build_config(key, value, d):
-            """ Called recursively to split up keys """
-            pieces = key.split('.', 1)
-            if len(pieces) == 1:
-                d[pieces[0]] = value.strip()
-            else:
-                d[pieces[0]] = _build_config(pieces[1], value, {})
-
-            return d
-
-        d = {}
-        for line in lines:
-            if '=' not in line:
-                continue
-
-            key, value = line.split('=', 1)
-            d = _build_config(key, value, d)
-
-        # Set a default data location if one is not specified.
-        if d.get('data') is None:
-            d['data'] = {}
-
-        if d['data'].get('location') is None:
-            d['data']['location'] = os.path.expanduser("~/.task/")
-
-        return d
+        return TaskRc(config_filename, overrides=overrides)
 
     @abc.abstractmethod
     def task_start(self, **kw):
@@ -314,6 +300,15 @@ class TaskWarriorDirect(TaskWarriorBase):
         if 'id' in task:
             del task['id']
 
+        # Delete None values (treat them as deleting values)
+        # https://github.com/ralphbean/taskw/pull/70
+        items = list(task.items())  # listify generator for py3 support.
+        for k, v in items:
+            if v is None:
+                task.pop(k)
+                if k in _task:
+                    _task.pop(k)
+
         _task.update(task)
         self._task_replace(line, Status.to_file(task['status']), _task)
         return line, _task
@@ -357,7 +352,6 @@ class TaskWarriorDirect(TaskWarriorBase):
         self._apply_modification(id, category, modification)
 
     def _apply_modification(self, id, category, modification):
-        location = self.config['data']['location']
         filename = DataFile.filename(category)
         filename = os.path.join(self.config['data']['location'], filename)
         filename = os.path.expanduser(filename)
@@ -412,27 +406,34 @@ class TaskWarriorShellout(TaskWarriorBase):
     and https://github.com/ralphbean/taskw/issues/30 for more.
     """
     DEFAULT_CONFIG_OVERRIDES = {
-        'json.array': 'TRUE',
+        'json': {
+            'array': 'TRUE'
+        },
         'verbose': 'nothing',
         'confirmation': 'no',
+        'dependency': {
+            'confirmation': 'no',
+        },
     }
 
-    def __init__(self, config_filename="~/.taskrc", config_overrides=None):
+    def __init__(
+        self,
+        config_filename=TASKRC,
+        config_overrides=None,
+        marshal=False,
+    ):
         super(TaskWarriorShellout, self).__init__(config_filename)
         self.config_overrides = config_overrides if config_overrides else {}
+        self._marshal = marshal
+        self.config = TaskRc(config_filename, overrides=config_overrides)
+
+        if self.get_version() >= LooseVersion('2.4'):
+            self.DEFAULT_CONFIG_OVERRIDES['verbose'] = 'new-uuid'
 
     def get_configuration_override_args(self):
-        args = []
         config_overrides = self.DEFAULT_CONFIG_OVERRIDES.copy()
         config_overrides.update(self.config_overrides)
-        for key, value in six.iteritems(config_overrides):
-            args.append(
-                'rc.%s=%s' % (
-                    key,
-                    value if ' ' not in value else '"%s"' % value
-                )
-            )
-        return args
+        return taskw.utils.convert_dict_to_override_args(config_overrides)
 
     def _execute(self, *args):
         """ Execute a given taskwarrior command with arguments
@@ -448,20 +449,48 @@ class TaskWarriorShellout(TaskWarriorBase):
             + self.get_configuration_override_args()
             + [six.text_type(arg) for arg in args]
         )
-        proc = subprocess.Popen(
-            command,
-            stdout=subprocess.PIPE,
-            stderr=subprocess.PIPE,
-        )
-        stdout, stderr = proc.communicate()
+
+        # subprocess is expecting bytestrings only, so nuke unicode if present
+        for i in range(len(command)):
+            if isinstance(command[i], six.text_type):
+                command[i] = command[i].encode('utf-8')
+
+        try:
+            proc = subprocess.Popen(
+                command,
+                stdout=subprocess.PIPE,
+                stderr=subprocess.PIPE,
+            )
+            stdout, stderr = proc.communicate()
+        except OSError as e:
+            if 'No such file or directory' in e:
+                raise OSError("Unable to find the 'task' command-line tool.")
+            raise
+
         if proc.returncode != 0:
-            raise TaskwarriorError(stderr, stdout, proc.returncode)
+            raise TaskwarriorError(command, stderr, stdout, proc.returncode)
+
+        # We should get bytes from the outside world.  Turn those into unicode
+        # as soon as we can.
+        stdout = stdout.decode(self.config.get('encoding', 'utf-8'))
+        stderr = stderr.decode(self.config.get('encoding', 'utf-8'))
+
         return stdout, stderr
 
     def _get_json(self, *args):
-        encoded = self._execute(*args)[0]
-        decoded = encoded.decode(self.config.get('encoding', 'utf-8'))
-        return json.loads(decoded)
+        return json.loads(self._execute(*args)[0])
+
+    def _get_task_objects(self, *args):
+        json = self._get_json(*args)
+        if isinstance(json, dict):
+            return self._get_task_object(json)
+        value = [self._get_task_object(j) for j in json]
+        return value
+
+    def _get_task_object(self, obj):
+        if self._marshal:
+            return Task(obj, udas=self.config.get_udas())
+        return obj
 
     def _stub_task(self, description, tags=None, **kw):
         """ Given a description, stub out a task dict. """
@@ -480,6 +509,9 @@ class TaskWarriorShellout(TaskWarriorBase):
 
         task.update(kw)
 
+        if self._marshal:
+            return Task.from_stub(task, udas=self.config.get_udas())
+
         return task
 
     @classmethod
@@ -495,10 +527,15 @@ class TaskWarriorShellout(TaskWarriorBase):
 
     @classmethod
     def get_version(cls):
-        taskwarrior_version = subprocess.Popen(
-            ['task', '--version'],
-            stdout=subprocess.PIPE
-        ).communicate()[0]
+        try:
+            taskwarrior_version = subprocess.Popen(
+                ['task', '--version'],
+                stdout=subprocess.PIPE
+            ).communicate()[0]
+        except OSError as e:
+            if 'No such file or directory' in e:
+                raise OSError("Unable to find the 'task' command-line tool.")
+            raise
         return LooseVersion(taskwarrior_version.decode())
 
     def sync(self, init=False):
@@ -515,7 +552,7 @@ class TaskWarriorShellout(TaskWarriorBase):
         """ Returns a dictionary of tasks for a list of command."""
 
         results = dict(
-            (db, self._get_json('status:%s' % db, 'export'))
+            (db, self._get_task_objects('status:%s' % db, 'export'))
             for db in Command.files(command)
         )
 
@@ -523,7 +560,7 @@ class TaskWarriorShellout(TaskWarriorBase):
         # Here we merge the waiting list back into the pending list.
         if 'pending' in results:
             results['pending'].extend(
-                self._get_json('status:waiting', 'export'))
+                self._get_task_objects('status:waiting', 'export'))
 
         return results
 
@@ -548,10 +585,8 @@ class TaskWarriorShellout(TaskWarriorBase):
         website.
 
         """
-        query_args = taskw.utils.encode_query(
-            filter_dict,
-        )
-        return self._get_json(
+        query_args = taskw.utils.encode_query(filter_dict, self.get_version())
+        return self._get_task_objects(
             'export',
             *query_args
         )
@@ -591,7 +626,7 @@ class TaskWarriorShellout(TaskWarriorBase):
             else:
                 search = [value]
 
-        task = self._get_json('export', *search)
+        task = self._get_task_objects('export', *search)
 
         if task:
             if isinstance(task, list):
@@ -612,11 +647,26 @@ class TaskWarriorShellout(TaskWarriorBase):
         # task and add them after we've added the task.
         annotations = self._extract_annotations_from_task(task)
 
-        task['uuid'] = str(uuid.uuid4())
-        stdout, stderr = self._execute(
-            'add',
-            taskw.utils.encode_task_experimental(task),
-        )
+        # With older versions of taskwarrior, you can specify whatever uuid you
+        # want when adding a task.
+        if self.get_version() < LooseVersion('2.4'):
+            task['uuid'] = str(uuid.uuid4())
+        elif 'uuid' in task:
+            del task['uuid']
+
+        if self._marshal:
+            args = taskw.utils.encode_task_experimental(task.serialized())
+        else:
+            args = taskw.utils.encode_task_experimental(task)
+
+        stdout, stderr = self._execute('add', *args)
+
+        # However, in 2.4 and later, you cannot specify whatever uuid you want
+        # when adding a task.  Instead, you have to specify rc.verbose=new-uuid
+        # and then parse the assigned uuid out from stdout.
+        if self.get_version() >= LooseVersion('2.4'):
+            task['uuid'] = stdout.strip().split()[-1].strip('.')
+
         id, added_task = self.get_task(uuid=task['uuid'])
 
         # Check if 'uuid' is in the task we just added.
@@ -641,6 +691,7 @@ class TaskWarriorShellout(TaskWarriorBase):
         self._execute(
             task['uuid'],
             'annotate',
+            '--',
             annotation
         )
         id, annotated_task = self.get_task(uuid=task[six.u('uuid')])
@@ -651,6 +702,7 @@ class TaskWarriorShellout(TaskWarriorBase):
         self._execute(
             task['uuid'],
             'denotate',
+            '--',
             annotation
         )
         id, denotated_task = self.get_task(uuid=task[six.u('uuid')])
@@ -671,7 +723,19 @@ class TaskWarriorShellout(TaskWarriorBase):
     def task_update(self, task):
         if 'uuid' not in task:
             raise KeyError('Task must have a UUID.')
-        id, original_task = self.get_task(uuid=task['uuid'])
+        # 'Legacy' causes us to handle this task as if it were an
+        # old-style task -- just a standard dictionary
+        legacy = True
+
+        if isinstance(task, Task):
+            # Let's pre-serialize taskw.task.Task instances
+            task_uuid = six.text_type(task['uuid'])
+            task = task.serialized_changes(keep=True)
+            legacy = False
+        else:
+            task_uuid = task['uuid']
+
+        id, original_task = self.get_task(uuid=task_uuid)
 
         if 'id' in task:
             del task['id']
@@ -681,35 +745,52 @@ class TaskWarriorShellout(TaskWarriorBase):
         task_to_modify.pop('uuid', None)
         task_to_modify.pop('id', None)
 
-        # Check if there are annotations, if so, look if they are
-        # in the existing task, otherwise annotate the task to add them.
-        ttm_annotations = taskw.utils.annotation_list_to_comparison_map(
-            self._extract_annotations_from_task(task_to_modify)
-        )
-        original_annotations = taskw.utils.annotation_list_to_comparison_map(
-            self._extract_annotations_from_task(original_task)
-        )
+        # Only handle annotation differences if this is an old-style
+        # task, or if the task itself says annotations have changed.
+        annotations_to_delete = set()
+        annotations_to_create = set()
+        if legacy or 'annotations' in task_to_modify:
+            # Check if there are annotations, if so, look if they are
+            # in the existing task, otherwise annotate the task to add them.
+            ttm_annotations = taskw.utils.annotation_list_to_comparison_map(
+                self._extract_annotations_from_task(task_to_modify)
+            )
+            original_annotations = (
+                taskw.utils.annotation_list_to_comparison_map(
+                    self._extract_annotations_from_task(original_task)
+                )
+            )
 
-        new_annotations = set(ttm_annotations.keys())
-        existing_annotations = set(original_annotations.keys())
+            new_annotations = set(ttm_annotations.keys())
+            existing_annotations = set(original_annotations.keys())
 
-        annotations_to_delete = existing_annotations - new_annotations
-        annotations_to_create = new_annotations - existing_annotations
+            annotations_to_delete = existing_annotations - new_annotations
+            annotations_to_create = new_annotations - existing_annotations
 
-        if 'annotations' in task_to_modify:
-            del task_to_modify['annotations']
+            if 'annotations' in task_to_modify:
+                del task_to_modify['annotations']
 
         modification = taskw.utils.encode_task_experimental(task_to_modify)
-        self._execute(task['uuid'], 'modify', modification)
+        # Only try to modify the task if there are changes to post here
+        # (changes *might* just be in annotations).
+        if modification:
+            self._execute(task_uuid, 'modify', *modification)
 
         # If there are no existing annotations, add the new ones
-        ttm_annotations.update(original_annotations)
-        for annotation_key in annotations_to_create:
-            self.task_annotate(original_task, ttm_annotations[annotation_key])
-        for annotation_key in annotations_to_delete:
-            self.task_denotate(original_task, ttm_annotations[annotation_key])
+        if legacy or annotations_to_delete or annotations_to_create:
+            ttm_annotations.update(original_annotations)
+            for annotation_key in annotations_to_create:
+                self.task_annotate(
+                    original_task,
+                    ttm_annotations[annotation_key]
+                )
+            for annotation_key in annotations_to_delete:
+                self.task_denotate(
+                    original_task,
+                    ttm_annotations[annotation_key]
+                )
 
-        return self.get_task(uuid=original_task['uuid'])
+        return self.get_task(uuid=task_uuid)
 
     def task_delete(self, **kw):
         """ Marks a task as deleted.  """
